import re
import shutil
import sys
from pathlib import Path
from datetime import datetime, timedelta
from typing import Iterator

from linc import write_nc_legacy
from linc.config import get_config, Config
from loguru import logger
from gfatpy.utils.io import unzip_file_to_temp

from gfatpy.utils.utils import parse_datetime
from ..file_manager import info2general_path, info2path
from ..types import LidarName, MeasurementType, Telescope
from .types import Measurement

CONFIGS_DIR = Path(__file__).parent / "configs"
DAYS_SEARCH_PAST = 5
RAW_FIRST_LETTERS = "RM"

logger.add(sys.stdout, level="INFO")


def convert_nc_by_date(
    date: datetime | str,
    lidar_name: LidarName = LidarName.mhc,
    measurement_type: MeasurementType | None = None,
    config_file: Path | str | None = None,
    data_dir: Path | str | None = None,
    output_dir: Path | str | None = None,
    telescope: Telescope = Telescope.xf,
) -> None:
    """Replaces the old lidar raw2l1 function

    Args:
        date (datetime | str): _description_
        lidar (LidarName, optional): _description_. Defaults to LidarName.mhc.
        measurement_type (MeasurementType | None, optional): _description_. Defaults to None.
        data_dir (Path | str | None, optional): _description_. Defaults to cwd.
    """
    if data_dir is None:
        data_dir = Path().cwd()
    elif isinstance(data_dir, str):
        data_dir = Path(data_dir)

    if not data_dir.exists() or not data_dir.is_dir():
        raise ValueError("Not found data_dir")

    if output_dir is None:
        output_dir = Path().cwd()
    elif isinstance(output_dir, str):
        output_dir = Path(output_dir)

    if not output_dir.exists() or not output_dir.is_dir():
        raise ValueError("Not found data_dir")

    if isinstance(config_file, str):
        config_file = Path(config_file)

    date = parse_datetime(date)
    config = search_config(lidar_name, config_file)
    origin_path = info2general_path(
        lidar_name.value, date=date, data_dir=data_dir
    )

    prev_paths = [
        info2general_path(
            lidar_name.value,
            date=date - timedelta(days=n_day),            
            data_dir=data_dir,
        )
        for n_day in range(1, DAYS_SEARCH_PAST + 1)
    ]

    measurements = to_measurements(origin_path.glob("*"))
    measurements = filter_by_type(measurements, measurement_type)

    if len(measurements) != 0:
        logger.info(f"{len(measurements)} measurements found")
    else:
        logger.error(f"0 measurements found")
        raise FileNotFoundError()

    rs_processed = False
    ot_processed = False

    for measurement in measurements:
        if measurement.type not in [MeasurementType.DP, MeasurementType.TC]:
            if measurement.type == MeasurementType.RS:
                if rs_processed:
                    continue
                files = glob_all_from_type_and_close_date(
                    measurements, MeasurementType.RS, prev_paths, date
                )
                rs_processed = True
            elif measurement.type == MeasurementType.OT:
                if ot_processed:
                    continue
                files = glob_all_from_type_and_close_date(
                    measurements, MeasurementType.OT, prev_paths, date
                )
                ot_processed = True
            else:                
                files = measurement.path.rglob(f"{RAW_FIRST_LETTERS}*")
            
            result_path = info2path(
                lidar_name=lidar_name.value,
                channel=get_532_from_telescope(telescope),
                measurement_type=measurement.type.value,
                date=measurement.time,
                dir=output_dir,
            )
            result_path.parent.mkdir(parents=True, exist_ok=True)
            logger.info(f"Writing {result_path.name}")            
            write_nc_legacy(files, result_path, config=config)
        else:
            sub_measurements = [*filter(lambda p: p.is_dir(), measurement.path.glob("*"))]
            sub_names = [path.name for path in sub_measurements]
            names_to_check = ['N45', 'P45', '+45', '-45', 'p45', 'n45']
            exist_in_list = any(name in sub_names for name in names_to_check)
            if not exist_in_list:
                sub_measurements = [*filter(lambda p: p.is_dir(), measurement.path.glob("*/*"))]

            for sm in sub_measurements:
                files = sm.rglob(f"{RAW_FIRST_LETTERS}*")
                result_path = info2path(
                    lidar_name=lidar_name.value,
                    channel=get_532_from_telescope(telescope),
                    measurement_type=measurement.type.value,
                    signal_type=sm.name,
                    date=measurement.time,
                    dir=output_dir,
                )
                result_path.parent.mkdir(parents=True, exist_ok=True)
                logger.info(f"Writing {result_path.name}")
                write_nc_legacy(files, result_path, config=config)
        
        # Find tmp folders matching the pattern to delete them
        #regular expression to find tmp folder
        pattern = r"tmp_unzipped_[a-zA-Z0-9]{8}$"
        if re.match(pattern, measurement.path.parent.name):
            shutil.rmtree(measurement.path.parent)
            print(f"Temporary folder deleted: {measurement.path.parent}")

def convert_nc_by_folder(
    dir: Path | str,
    date: datetime | str,
    output_dir: Path | str | None = None,
    lidar_name: LidarName = LidarName.mhc,
    measurement_type: MeasurementType | None = None,
    config_file: str | None = None,
    telescope: Telescope = Telescope.xf,
) -> None:

    if isinstance(dir, str):
        dir = Path(dir)

    if not dir.exists() or not dir.is_dir():
        raise ValueError("Not found data_dir")

    if isinstance(output_dir, str):
        output_dir = Path(output_dir)
    elif output_dir is None:
        output_dir = Path.cwd()

    if not output_dir.exists() or not output_dir.is_dir():
        raise ValueError("Not found output_dir")

    date = parse_datetime(date)
    config = search_config(lidar_name, config_file)
    origin_path = dir

    measurements = to_measurements(origin_path.glob("*"))
    measurements = filter_by_type(measurements, measurement_type)
    if len(measurements) != 0:
        logger.info(f"{len(measurements)} measurements found")
    else:
        logger.error(f"0 measurements found")
        raise FileNotFoundError()

    rs_processed = False
    ot_processed = False

    for measurement in measurements:
        if measurement.type not in [MeasurementType.DP, MeasurementType.TC]:
            if measurement.type == MeasurementType.RS:

                if rs_processed:
                    continue
                files = glob_all_from_type_and_close_date(
                    measurements, MeasurementType.RS, [], date
                )
                rs_processed = True
            if measurement.type == MeasurementType.OT:
                if ot_processed:
                    continue
                files = glob_all_from_type_and_close_date(
                    measurements, MeasurementType.OT, [], date
                )
                ot_processed = True
            else:
                files = measurement.path.glob(f"{RAW_FIRST_LETTERS}*")
            result_path = info2path(
                lidar_name=lidar_name.value,
                channel=get_532_from_telescope(telescope),
                measurement_type=measurement.type.value,
                data_level="1a",
                date=measurement.time,
                dir=output_dir,
            )
                        
            result_path.parent.mkdir(parents=True, exist_ok=True)
            logger.info(f"Writing {result_path.name}")
            write_nc_legacy(files, result_path, config=config)
        else:
            sub_measurements = filter(lambda p: p.is_dir(), measurement.path.glob("*"))
            for sm in sub_measurements:
                files = sm.glob(f"{RAW_FIRST_LETTERS}*")
                result_path = info2path(
                    lidar_name=lidar_name.value,
                    channel=get_532_from_telescope(telescope),
                    measurement_type=measurement.type.value,
                    data_level="1a",
                    date=measurement.time,
                    dir=output_dir,
                )
                result_path.parent.mkdir(parents=True, exist_ok=True)
                logger.info(f"Writing {result_path.name}")
                write_nc_legacy(files, result_path, config=config)


def search_config(
    lidar_name: LidarName, opt_config: Path | str | None = None
) -> Config:

    try_last_chance = False
    if isinstance(opt_config, Path):
        config_path = opt_config
    elif isinstance(opt_config, str):
        config_path = Path(opt_config)
        if not config_path.exists():
            try_last_chance = True
    elif opt_config is None:
        try_last_chance = True

    if try_last_chance:
        config_path = Path(CONFIGS_DIR / f"{lidar_name.value.upper()}.toml")
    else:
        raise FileNotFoundError(f"No configution file found in {opt_config}")

    if not config_path.exists():
        raise FileNotFoundError(f"No configution file found in {opt_config}")

    return get_config(config_path)


def filter_by_type(
    measurements: list[Measurement], mtype: MeasurementType | None = None
) -> list[Measurement]:
    match mtype:
        case MeasurementType.RS:
            measurements = list(
                filter(
                    lambda m: m.type in [MeasurementType.RS, MeasurementType.DC],
                    measurements,
                )
            )
        case MeasurementType.TC:
            measurements = list(
                filter(
                    lambda m: m.type in [MeasurementType.TC, MeasurementType.DC],
                    measurements,
                )
            )
        case MeasurementType.DP:
            measurements = list(
                filter(
                    lambda m: m.type in [MeasurementType.DP, MeasurementType.DC],
                    measurements,
                )
            )
        case MeasurementType.HF:
            measurements = list(
                filter(
                    lambda m: m.type in [MeasurementType.HF, MeasurementType.DC],
                    measurements,
                )
            )
        case MeasurementType.OT:
            measurements = list(
                filter(lambda m: m.type in [MeasurementType.OT], measurements)
            )
        case MeasurementType.DC:
            measurements = list(
                filter(lambda m: m.type in [MeasurementType.DC], measurements)
            )
        case None:
            return measurements
    return measurements


def glob_all_from_type_and_close_date(
    measurements: list[Measurement],
    mtype: MeasurementType,
    prev_paths: list[Path],
    date: datetime,
) -> list[Path]:
    paths: set[Path] = set({})

    # Group all groups of measurements into one array
    for meas in measurements:
        if meas.type == mtype:
            paths |= set(meas.path.rglob(f"{RAW_FIRST_LETTERS}*"))

    # Aditionally, search in neighbor days from measurements of this day
    for prev_path in prev_paths:
        prev_measurements = prev_path.glob(f"{mtype.value}_*")
        for prev_measurement in prev_measurements:
            same_day_measurements = set(
                prev_measurement.rglob(f"{RAW_FIRST_LETTERS}{to_licel_date_str(date)}*")
            )
            paths |= same_day_measurements  # Set union assignment

    return list(paths)


def to_measurements(glob: Iterator[Path]) -> list[Measurement]:
    measurements = []
    for path in glob:        
        if path.suffix == ".zip":
            #Unzip file into a temporary folder
            meas_path  = unzip_file_to_temp(path)            
        else:
            meas_path = path
        time = datetime.strptime(meas_path.name[3:], r"%Y%m%d_%H%M")
        measurements.append(
            Measurement(type=MeasurementType(path.name[:2]), path=meas_path, time=time)
        )    
    return measurements


def get_532_from_telescope(telescope: Telescope = Telescope.xf) -> str:
    if telescope == telescope.xf:
        return "532xpa"
    elif telescope == telescope.ff:
        return "532fpa"
    elif telescope == telescope.nf:
        return "532npa"

    raise ValueError("Telescope type not recognized. Options are xf, ff, nf")


def to_licel_date_str(date: datetime) -> str:
    month_hex = f"{date.month:x}"
    return f'{date.strftime(r"%y")}{month_hex}{date.strftime(r"%d")}'



from pathlib import Path
import matplotlib.pyplot as plt
import xarray as xr

from gfatpy.lidar.utils import signal_to_rcs


def plot_eta_start_calib(
    d: xr.Dataset,
    wavelength: int = 532,
    output_dir: Path | str | None = None,
    telescope: str = "x",
) -> Path | None:
    """_summary_

    Args:
        d (xr.Dataset): Eta star dataset
    """
    if isinstance(output_dir, str):
        output_dir = Path(output_dir)
    elif output_dir is None:
        output_dir = Path.cwd()

    if not output_dir.exists() or not output_dir.is_dir():
        raise NotADirectoryError(f"{output_dir} not found.")

    try:
        channel_str = f"{wavelength}{telescope}"

        # FIXME: This search channels should be provisional and shoud be make by
        # passing a channels array with the eta star dataset
        variables: list[str] = list(d.variables.keys())  # type: ignore
        search_channels = filter(lambda v: v.startswith("eta_star_profile_"), variables)
        channel_names = list(map(lambda v: v[17:-1], search_channels))

        fig, (ax1, ax2, ax3, ax4) = plt.subplots(1, 4, figsize=(15, 10), sharey=True)
        y_lim = (0, 12_000)
        # Title
        fig_title_1 = r"Calibration depolarization analysis"
        # % (time_start_dt.strftime("%Y%m%d"), time_start_dt.strftime("%H%M"),
        #     time_end_dt.strftime("%H%M"))
        # fig_title_2 = r"Channel %s. $\alpha_{misalign.}$ ($\circ$)= %4.2f $\pm$ %4.2f" \
        #                 % (wv_str, dw["alpha_%s" % ch_an], dw["alpha_error_%s" % ch_an])
        # fig_title_3 = r"Analog: Calib. factor [%3.1f - %3.1f km] = %6.4f $\pm$ %6.4f ; " \
        #                 r"$\epsilon_{misalig.}$ ($^\circ$) = %4.2f $\pm$ %4.2f" \
        #                 % (dw["cal_height_an"][0]*1e-3, dw["cal_height_an"][1]*1e-3,
        #                     dw["gain_ratio_Delta90_avg_%s" % ch_an], dw["gain_ratio_Delta90_std_%s" % ch_an],
        #                     dw["epsilon_%s" % ch_an], dw["epsilon_error_%s" % ch_an])
        # fig_title_4 = r"Photoncounting: Calib. factor [%3.1f - %3.1f km] = %6.4f $\pm$ %6.4f ; " \
        #                 r"$\epsilon_{misalig.}$ ($^\circ$) = %4.2f $\pm$ %4.2f" \
        #                 % (dw["cal_height_pc"][0]*1e-3, dw["cal_height_pc"][1]*1e-3,
        #                     dw["gain_ratio_Delta90_avg_%s" % ch_pc], dw["gain_ratio_Delta90_std_%s" % ch_pc],
        #                     dw["epsilon_%s" % ch_pc], dw["epsilon_error_%s" % ch_pc])

        plt.suptitle(f"{fig_title_1} \n")

        # ANALOG
        signal_to_rcs(d[f"signal_T_P45_{channel_str}a"], d.range).plot(
            ax=ax1, y="range", lw=2, c="lime", label=r"RCS$^T_{+45}$. AN"
        )  # type: ignore
        signal_to_rcs(d[f"signal_R_N45_{channel_str}a"], d.range).plot(
            ax=ax1, y="range", lw=2, c="darkgreen", label=r"RCS$^R_{-45}$. AN"
        )  # type: ignore
        signal_to_rcs(d[f"signal_T_N45_{channel_str}a"], d.range).plot(
            ax=ax1, y="range", lw=2, c="r", label=r"RCS$^T_{-45}$. AN"
        )  # type: ignore
        signal_to_rcs(d[f"signal_R_P45_{channel_str}a"], d.range).plot(
            ax=ax1, y="range", lw=2, c="darkred", label=r"RCS$^R_{+45}$. AN"
        )  # type: ignore

        # PHOTONCOUNTING
        signal_to_rcs(d[f"signal_T_P45_{channel_str}p"], d.range).plot(ax=ax1, y="range", lw=2, c="deepskyblue", label=r"RCS$^T_{+45}$. PC")  # type: ignore
        signal_to_rcs(d[f"signal_R_N45_{channel_str}p"], d.range).plot(
            ax=ax1, y="range", lw=2, c="b", label=r"RCS$^R_{-45}$. PC"
        )  # type: ignore
        signal_to_rcs(d[f"signal_T_N45_{channel_str}p"], d.range).plot(
            ax=ax1, y="range", lw=2, c="magenta", label=r"RCS$^T_{-45}$. PC"
        )  # type: ignore
        signal_to_rcs(d[f"signal_R_P45_{channel_str}p"], d.range).plot(
            ax=ax1, y="range", lw=2, c="darkmagenta", label=r"RCS$^R_{+45}$. PC"
        )  # type: ignore

        ax1.grid()
        ax1.axes.set_xlabel(r"RCS [a.u.]")
        ax1.axes.set_ylabel(r"Height [km, agl]")
        ax1.axes.set_ylim(y_lim)  # type: ignore
        ax1.axes.set_xscale("log")
        ax1.legend(fontsize="small", loc=2)

        (d[f"signal_R_P45_{channel_str}a"] / d[f"signal_T_P45_{channel_str}a"]).plot(
            ax=ax2, y="range", lw=2, c="blue", label=r"AN at +45"
        )  # type: ignore
        (d[f"signal_R_N45_{channel_str}a"] / d[f"signal_T_N45_{channel_str}a"]).plot(
            ax=ax2, y="range", lw=2, c="red", label=r"AN at -45"
        )  # type: ignore
        (d[f"signal_R_P45_{channel_str}p"] / d[f"signal_T_P45_{channel_str}p"]).plot(
            ax=ax2, y="range", lw=2, c="green", label=r"PC at +45"
        )  # type: ignore
        (d[f"signal_R_N45_{channel_str}p"] / d[f"signal_T_N45_{channel_str}p"]).plot(
            ax=ax2, y="range", lw=2, c="black", label=r"PC at -45"
        )  # type: ignore

        ax2.grid()
        ax2.axes.set_xlabel(r"Signal Ratio [R / T]")
        ax2.axes.set_ylabel(r"")
        ax2.legend(fontsize="small", loc=7)
        ax2.axes.set_xlim((0, 1))

        d[f"eta_star_profile_{channel_str}a"].plot(
            ax=ax3, y="range", c="g", label=r"AN"
        )  # type: ignore
        d[f"eta_star_profile_{channel_str}p"].plot(
            ax=ax3, y="range", c="m", label=r"PC"
        )  # type: ignore
        ax3.grid()
        ax3.axes.set_xlabel(r"Calibration Factor")
        ax3.axes.set_ylabel(r"")
        ax3.axes.set_xlim((0, 10))
        ax3.legend(fontsize="small", loc=7)
        # ax3.axes.set_xlim((0, 1))
    except ValueError:
        output_file = None
        raise ValueError

    if output_dir.exists() and output_dir.is_dir():
        calibration_datetime = d.attrs["calibration_datetime"]
        output_file = (
            output_dir / f"calibration_{channel_str}_{calibration_datetime}.png"
        )
        plt.savefig(output_file, dpi=300, bbox_inches="tight")
    else:
        output_file = None
        raise FileNotFoundError("`output_dir` not found.")
    return output_file



#!/usr/bin/env python
import os
import glob
import pathlib
import warnings

import datetime as dt
import numpy as np
import xarray as xr
from loguru import logger

from gfatpy import utils
from gfatpy.lidar.utils import LIDAR_INFO, get_lidar_name_from_filename

warnings.filterwarnings("ignore")


__author__ = "Bravo-Aranda, Juan Antonio"
__license__ = "GPL"
__version__ = "1.0"
__maintainer__ = "Bravo-Aranda, Juan Antonio"
__email__ = "jabravo@ugr.es"
__status__ = "Production"

""" DEFAULT AUXILIAR INFO
"""
# Root Directory (in NASGFAT)  according to operative system


def reader_xarray(
    filelist: list[str] | str,
    date_ini: str | None = None,
    date_end: str | None = None,
    ini_range: float | None = None,
    end_range: float | None = None,
    percentage_required: int = 80,
    channels: list[str] = [],
) -> xr.Dataset:
    """
    Lidar data reader using xarray module.
    Inputs:
    - filelist: String with pattern for create a List of lidar files (i.e, '/drives/c/*.nc') (str)
                List of lidar files (list)
    - date_ini: 'yyyymmddThhmmss'
    - date_end: 'yyyymmddThhmmss'
    - ini_range: int/float (m)
    - end_range: int/float (m)
    - percentage_required= percentage of the time period required to continue the process. Default 80%  (int)
    - channels: list of channel number (e.g., [0, 1, 5]) or [] to load all of them
    Output:
    - lidar: dictionary or 'None' in case of error.
    """

    """ Aux Functions
    """

    def add_required_channels(
        lidar_name: str,
        channels: list | np.ndarray,
    ) -> list:

        if isinstance(channels, np.ndarray):
            channels = channels.tolist()

        required_channels = channels.copy()
        for channel_ in channels:
            if channel_ in LIDAR_INFO["lidars"][lidar_name]["product_channels"].keys():
                required_channels = required_channels + [
                    *LIDAR_INFO["lidars"][lidar_name]["product_channels"][channel_]
                ]
        unique_channels = np.unique(np.array(required_channels)).tolist()
        return unique_channels

    def select_channels(dataset: xr.Dataset, channels: list | str) -> xr.Dataset:
        """select_channels function creates a new dataset with 'signal_CHANNEL' defined in 'channels' (list).

        Args:
            dataset (xr.Dataset): lidar dataset
            channels (list | str): list of lidar channel names

        Returns:
            xr.Dataset: lidar dataset
        """

        if len(channels) > 0:
            if isinstance(channels, str):
                channels = [channels]
            if isinstance(channels, np.ndarray):
                channels = channels.tolist()

            # find variables: signal related to channel
            _vars = ["signal"]
            real_channels = []
            for _channel in dataset["channel"].values.tolist():
                if _channel not in channels:
                    for _var in _vars:
                        varname = "%s_%s" % (_var, _channel)
                        dataset = dataset.drop_vars(varname)
                else:
                    real_channels.append(_channel)
            dataset = dataset.sel(channel=real_channels)
            dataset = dataset.assign_coords(channel=real_channels)
        return dataset

    def check_minimum_profiles(
        times: np.ndarray,
        date_ini: dt.datetime,
        date_end: dt.datetime,
        percentage_required: float,
    ) -> bool:
        """Check Lidar Data has enough profiles

        Args:
            times ([type]): [description]
            date_ini ([type]): [description]
            date_end ([type]): [description]
            percentage_required ([type]): [description]
        """

        check = True
        time_resolution = float(
            np.median(np.diff(times)) / np.timedelta64(1, "s")
        )  # FIXME: typing error
        interval_duration = (date_end - date_ini).total_seconds()
        Nt = np.round(
            interval_duration / time_resolution
        )  # Theoretical Number of profiles
        Nm = (percentage_required / 100) * Nt  # Minimum number of profiles
        Np = len(times)  # Number of actual profiles
        if Np > Nm:
            logger.info(
                f"Data loaded from {date_ini.isoformat()} to {date_end.isoformat()}"
            )
        else:
            logger.warning(
                f"Not enough data found ({Np}<{Nm}) in the user-required period ({interval_duration} s.)"
            )
            check = False

        return check

    """ The Reader
        The Reader does:
        1. concatenate along time dimension
        2. merge channels comming from different telescopes (ALHAMBRA), assuming same range coordinate
    """
    logger.info("Start Reader ...")

    # Find Files to Read
    try:
        if isinstance(filelist, str):
            filelist = [filelist]
        elif isinstance(filelist, pathlib.Path):
            filelist = [str(filelist.absolute())]
        files2load = []
        for i_file in filelist:
            files2load.extend(glob.glob(i_file))
    except Exception as e:
        files2load = []
        logger.warning(str(e))
        logger.warning(f"Files in {filelist} not found.")
    if len(files2load) > 0:
        logger.info(files2load)
        lidartemp = None

        lidar_nick = get_lidar_name_from_filename(files2load[0])
        if lidar_nick is None:
            logger.critical("Lidar nick not in lidar systems availables.")

        try:
            # Get lidar name
            lidar_name = LIDAR_INFO["metadata"]["nick2name"][lidar_nick]

            # Add required channels to the channel list to obtain product channel:
            channels = add_required_channels(lidar_name, channels)
            # Loop over modules: 1) concat time; 2) merge module
            for module in LIDAR_INFO["lidars"][lidar_name]["modules"]:
                module_fns = [x for x in files2load if module in x]
                lidarmod = None
                if len(module_fns) > 0:
                    for fn in module_fns:
                        with xr.open_dataset(
                            fn, chunks={}
                        ) as _dx:  # chunks={"time": 600, "range": 1000})
                            _dx = select_channels(_dx, channels)
                        if not lidarmod:
                            lidarmod = _dx
                        else:
                            # concat only variables that have "time" dimension.
                            # rest of variables keep values from first dataset
                            try:
                                lidarmod = xr.concat(
                                    [lidarmod, _dx],
                                    dim="time",
                                    data_vars="minimal",
                                    coords="minimal",
                                    compat="override",
                                )
                            except Exception as e:
                                logger.critical("Dataset in {fn} not concatenated")
                                raise e
                    # Sort Dataset by Time
                    if lidarmod is not None:
                        lidarmod = lidarmod.sortby(lidarmod["time"])
                    else:
                        raise ValueError("lidarmod is None.")
                    # Merge Module
                    if not lidartemp:
                        lidartemp = lidarmod
                    else:
                        try:
                            lidartemp = xr.merge([lidartemp, lidarmod])
                        except Exception as e:
                            logger.critical(f"{e}")
                            logger.critical(f"Dataset from module {module} not merged")
                            raise e
                del lidarmod
        except Exception as e:
            logger.critical(f"{e}")
            logger.critical("Files not concatenated")
            raise e

        if lidartemp:
            # Selection time window and Check Enough Profiles
            if np.logical_and(date_ini is not None, date_end is not None):
                if np.logical_and(isinstance(date_ini, str), isinstance(date_end, str)):
                    # Times Formatting

                    date_ini_dt = utils.utils.str_to_datetime(date_ini)  # type: ignore
                    date_end_dt = utils.utils.str_to_datetime(date_end)  # type: ignore

                    # Time Selection
                    min_time_resol = dt.timedelta(seconds=0.1)
                    lidar = lidartemp.sel(
                        time=slice(
                            date_ini_dt - min_time_resol, date_end_dt + min_time_resol
                        )
                    )

                    # Check selection
                    ok = check_minimum_profiles(
                        lidar["time"].values,
                        date_ini_dt,
                        date_end_dt,
                        percentage_required,
                    )
                    if not ok:
                        lidar = None
                else:
                    lidar = lidartemp
            else:
                lidar = lidartemp
            del lidartemp

            # Complete lidar dataset
            if lidar:
                # Range Clip
                if ini_range is not None and end_range is not None:
                    if end_range > ini_range:
                        lidar = lidar.sel(range=slice(ini_range, end_range))
                    else:
                        raise ValueError("ini_range is larger than end_range.")

                # add background ranges
                if "BCK_MIN_ALT" not in lidar.attrs.keys():
                    lidar.attrs["BCK_MIN_ALT"] = 75000
                if "BCK_MAX_ALT" not in lidar.attrs.keys():
                    lidar.attrs["BCK_MAX_ALT"] = 105000

                # Extract information from filename
                try:
                    lidar.attrs["lidarNick"] = os.path.basename(files2load[0]).split(
                        "_"
                    )[0]
                    lidar.attrs["dataversion"] = os.path.basename(files2load[0]).split(
                        "_"
                    )[1]
                except:
                    lidar.attrs["lidarNick"] = "Unknown"
                    lidar.attrs["dataversion"] = "Unknown"

            else:
                lidar = None
        else:
            logger.error("Impossible to load found files.")
            lidar = None
    else:
        logger.error("Files not found.")
        lidar = None

    if lidar is None:
        # set_trace()
        logger.error("No dataset created.")
        raise RuntimeError("No dataset created.")

    logger.info("End Reader")
    return lidar



# from . import preprocessing, retrieval, plot, utils


# __all__ = ["preprocessing", "retrieval", "plot", "utils"]


# __doc__ = """
#     The top of the radar module that is compatible (at least) with GFAT radar: "NEPHELE" AND "NEBULA"
# """



from pathlib import Path
from typing import Any, Dict, Union

import numpy as np
import xarray as xr
from datetime import datetime

from gfatpy.lidar import file_manager
from gfatpy.lidar.utils import LIDAR_INFO
from gfatpy.lidar.depolarization import io
from gfatpy.utils.io import find_nearest_filepath

CDIR = Path(__file__).parent.parent.parent.parent


def backscattering_ratio(
    molecular_backscatter: np.ndarray, particle_backscatter: np.ndarray
) -> np.ndarray[Any, np.dtype[np.float64]]:
    """Retrieves the backscattering ratio. Inputs must be in the same units.

    Args:
        molecular_backscatter (np.ndarray): Molecular backscatter coefficient.
        particle_backscatter (np.ndarray):  Particle backscatter coefficient.

    Returns:
        np.ndarray: backscattering ratio
    """

    return molecular_backscatter + particle_backscatter / molecular_backscatter


def linear_volume_depolarization_ratio(
    signal_R: np.ndarray,
    signal_T: np.ndarray,
    channel_R: str,
    channel_T: str,
    range: np.ndarray,
    time: np.ndarray,
    eta_star: float = 1,
    K: float = 1,
    GT: float = 1,
    HT: float = -1,
    GR: float = 1,
    HR: float = 1,
) -> xr.DataArray:
    """Calculate the linear volume depolarization ratio.

    Args:
        signal_R (np.ndarray): reflected signal in the polarizing beam splitter cube.
        signal_T (np.ndarray): transmitted signal in the polarizing beam splitter cube.
        channel_R (str): reflected channel in the polarizing beam splitter cube.
        channel_T (str): transmitted channel in the polarizing beam splitter cube.
        range (np.ndarray): range series of the signal.
        time (np.ndarray): time vector of the signal.
        eta_star (float, optional): calibration factor retrieved with Delta90 method. Defaults to 1.
        K (float, optional): K factor value simulated with Volker's algorithm. Defaults to 1.
        GT (float, optional): GT factor value simulated with Volker's algorithm. Defaults to 1.
        HT (float, optional): HT factor value simulated with Volker's algorithm. Defaults to -1.
        GR (float, optional): GR factor value simulated with Volker's algorithm. Defaults to 1.
        HR (float, optional): HR factor value simulated with Volker's algorithm. Defaults to 1.
        Y (float, optional): Optical system orientation (laser-beam Vs beam-splitter cube).

    Raises:
        ValueError: Wavelength, telescope or mode does not fit raises 'Polarized channels to be merged not appropiated'.
        ValueError: polarization codes do not fit raises 'Polarized channels to be merged not appropiated'.

    Returns:
        xr.DataArray: linear_volume_depolarization_ratio
    """

    wavelengthR, telescopeR, polR, modeR = file_manager.channel2info(channel_R)
    wavelengthT, telescopeT, polT, modeT = file_manager.channel2info(channel_T)

    if wavelengthR != wavelengthT or telescopeR != telescopeT or modeR != modeT:
        raise ValueError("Polarized channels to be merged not appropiated.")

    if polT not in ["p", "c", "s"] or polR not in ["p", "c", "s"] and polR != polT:
        raise ValueError("Polarized channels to be merged not appropiated.")

    eta = eta_star / K
    ratio = (signal_R / signal_T) / eta
    lvdr_ = (((GT + HT) * ratio - (GR + HR)) / ((GR - HR) - (GT - HT) * ratio)).astype(
        float
    )

    # Create DataArray
    lvdr = xr.DataArray(
        lvdr_,
        coords={"time": time, "range": range},
        dims=["time", "range"],
        attrs={
            "long_name": "Linear Volume Depolarization Ratio",
            "detection_mode": LIDAR_INFO["metadata"]["code_mode_str2number"][modeR],
            "wavelength": wavelengthR,
            "units": "$\\#$",
        },
    )
    return lvdr


def add_depolarization_products_of_channel(
    dataset: xr.Dataset,
    channel_depo: str,
    ghk: Dict[str, Union[float, None]],
    calibration: xr.Dataset,
) -> xr.Dataset:
    """Add total (parallel + cross) signal and the linear volume depolarization ratio for the polarizing channels to the `lidar dataset`.

    Args:
        dataset (xr.Dataset): Lidar dataset from `lidar.preprocess`.
        channel_depo (str): Depolarization channel [e.g., '532x', '355n'].
        ghk (Dict[str, Union[float, None]]): `channel_depo` GHK values from `gfatpy.lidar.depolarization.io.ghk_output_reader`.
        calibration (xr.Dataset): Depolarization calibration dataset from `gfatpy.lidar.depolarization.io.eta_star_reader`.

    Raises:
        ValueError: At least one GHK value is None.
        KeyError: `channel_depo` eta_star_mean is not in the calibration dataset.

    Returns:
        xr.Dataset: Given lidar dataset with the depolarization products.
    """

    # Check GHK values are not None
    if "K2" in ghk.keys():
        if ghk["K2"] is not None:
            K = ghk["K2"]
        else:
            raise ValueError("K is None.")
    else:
        raise KeyError("K is not in GHK values.")
    if "GT" in ghk.keys():
        if ghk["GT"] is not None:
            GT = ghk["GT"]
        else:
            raise ValueError("GT is None.")
    else:
        raise KeyError("GT is not in GHK values.")
    if "HT" in ghk.keys():
        if ghk["HT"] is not None:
            HT = ghk["HT"]
        else:
            raise ValueError("HT is None.")
    else:
        raise KeyError("HT is not in GHK values.")
    if "GR" in ghk.keys():
        if ghk["GR"] is not None:
            GR = ghk["GR"]
        else:
            raise ValueError("GR is None.")
    else:
        raise KeyError("GR is not in GHK values.")
    if "HR" in ghk.keys():
        if ghk["HR"] is not None:
            HR = ghk["HR"]
        else:
            raise ValueError("HR is None.")
    else:
        raise KeyError("HR is not in GHK values.")

    # Get lidar name
    lidar_name: str = dataset.attrs["system"].lower()

    # Get channel info
    wavelength_, telescope_ = int(channel_depo[0:-1]), channel_depo[-1]

    # Get polarization channels to be processed
    polarized_channels = LIDAR_INFO["lidars"][lidar_name]["polarized_channels"]

    for mode_ in polarized_channels[f"{telescope_}f"][wavelength_].keys():
        # Data
        channel_R = polarized_channels[f"{telescope_}f"][wavelength_][mode_]["R"]
        channel_T = polarized_channels[f"{telescope_}f"][wavelength_][mode_]["T"]

        if (
            dataset.get(f"signal_{channel_R}") is not None
            and dataset.get(f"signal_{channel_T}") is not None
        ):
            try:
                eta_star = calibration[
                    f"eta_star_mean_{wavelength_}{telescope_}{mode_}"
                ].item()
            except KeyError:
                raise KeyError(
                    f"eta_star_mean_{wavelength_}{telescope_}{mode_} must be a key in the calibration xr.dataset."
                )

            # Sum of parallel and cross polarized channels
            dataset[f"signal_{wavelength_}{telescope_}t{mode_}"] = (
                eta_star * HR * dataset[f"signal_{channel_T}"]
                - HT * dataset[f"signal_{channel_R}"]
            )

            # Retrieve LVDR
            signal_R = dataset[f"signal_{channel_R}"].values
            signal_T = dataset[f"signal_{channel_T}"].values
            ranges = dataset.range.values
            times = dataset.time.values

            lvdr = linear_volume_depolarization_ratio(
                signal_R,
                signal_T,
                channel_R,
                channel_T,
                ranges,
                times,
                eta_star,
                K,
                GT,
                HT,
                GR,
                HR,
            )

            # Save lvdr in dataset
            key_channel = channel_R.replace("pa", "a").replace("pp", "p")
            dataset[f"linear_volume_depolarization_ratio_{key_channel}"] = lvdr

    return dataset


def add_depolarization_products(
    dataset: xr.Dataset, ghk_dir: Path, depo_calib_dir: Path
) -> xr.Dataset:

    # Check directories exist
    if not ghk_dir.exists() and not ghk_dir.exists():
        raise FileNotFoundError(f"{ghk_dir} does not exist.")
    if not depo_calib_dir.exists() and not depo_calib_dir.exists():
        raise FileNotFoundError(f"{depo_calib_dir} does not exist.")

    # Get lidar name
    lidar_name: str = dataset.attrs["system"].lower()
    lidar_nick: str = LIDAR_INFO["metadata"]["name2nick"][lidar_name]

    # Get datetime from dataset
    target_date: datetime = dataset.time.values[0]

    # Get polarization channels to be processed
    polarized_channels = LIDAR_INFO["lidars"][lidar_name]["polarized_channels"]
    for telescope_ in polarized_channels.keys():
        for wavelength_ in polarized_channels[telescope_].keys():
            for mode_ in polarized_channels[telescope_][wavelength_].keys():

                # Channel depo
                depo_channel = f"{wavelength_}{telescope_[0]}"

                # Get GHK file
                ghk_file = find_nearest_filepath(
                    ghk_dir,
                    f"*{lidar_nick}*{wavelength_}{telescope_[0]}*.dat",
                    6,
                    target_date,
                )
                ghk = io.ghk_output_reader(ghk_file)

                # Get depolarization calibration file
                depo_calib_file = find_nearest_filepath(
                    depo_calib_dir, f"*{lidar_nick}*eta-star*.nc", 2, target_date
                )
                depo_calib = xr.open_dataset(depo_calib_file)

                # Add depolarization products
                dataset = add_depolarization_products_of_channel(
                    dataset, depo_channel, ghk, depo_calib
                )
    return dataset


def search_ghk(file_list: list, date: datetime) -> Path:
    # FIXME: to be done.
    ghk_filepath = Path()
    return ghk_filepath


def particle_depolarization(
    linear_volume_depolarization_ratio: np.ndarray,
    backscattering_ratio: np.ndarray,
    molecular_depolarization: float,
    time: np.ndarray,
    range: np.ndarray,
) -> xr.DataArray:  # FIXME: this function may realy direction to the lvdr xr.DataArray inside the lidar dataset once executed lvdr_from_dataset.
    """Calculate the linear particle depolarization ratio.

    Args:
        linear_volume_depolarization_ratio (np.ndarray): linear volume depolarization ratio
        backscattering_ratio (np.ndarray): _description_
        molecular_depolarization (float): molecular linear volume depolarization ratio
        time (np.ndarray): time vector of the signal.
        range (np.ndarray): range series of the signal.

    Returns:
        xr.DataArray: linear particle depolarization ratio.

    Notes:
    The linear particle depolarization ratio is calculated by the formula:

    .. math::
       \\delta^p = \\frac{(1 + \\delta^m)\\delta^V \\mathbf{R} - (1 + \\delta^V)\\delta^m}
       {(1 + \\delta^m)\\mathbf{R} - (1 + \\delta^V)}


    References:
    Freudenthaler, V. et al. Depolarization ratio profiling at several wavelengths in pure
    Saharan dust during SAMUM 2006. Tellus, 61B, 165-179 (2008)
    """

    delta_p = (
        (1 + molecular_depolarization)
        * linear_volume_depolarization_ratio
        * backscattering_ratio
        - (1 + linear_volume_depolarization_ratio) * molecular_depolarization
    ) / (
        (1 + molecular_depolarization) * backscattering_ratio
        - (1 + linear_volume_depolarization_ratio)
    )

    # Create DataArray
    delta_p_xarray = xr.DataArray(
        delta_p,
        coords={"time": time, "range": range},
        dims=["time", "range"],
        attrs={
            "long_name": "Linear Volume Depolarization Ratio",
            "units": "$\\#$",
        },  # TODO: decide where the attributes should be included.
    )
    return delta_p_xarray



import os

import xarray as xr
import netCDF4 as nc
import numpy as np
from loguru import logger


def cloudnet_reader(cloudnet_fn: str, fields: list | None = None) -> xr.Dataset:
    """Read Cloudnet File

    Parameters
    ----------
    cloudnet_fn: str
        full path of cloudnet file name
    fields: list
        list of variables within the cloudnet dataset

    Returns
    -------
    ds: xarray
        Cloudnet dataset
    """
    try:
        if os.path.isfile(cloudnet_fn):
            with xr.open_dataset(cloudnet_fn, chunks={}) as ds:
                if fields is not None:
                    ds = ds[fields]
        else:
            logger.error("File %s not found" % cloudnet_fn)
            ds = None
    except Exception as e:
        logger.error(str(e))
        ds = None

    return ds


def categorice_reader(list_files: list[str]) -> dict:
    """read categorice data from netCDF files

    Designed to work with netcdf list_files associated to one single day
    """
    logger.warning("to be deprecated")
    print("Reading netCDF files: %s" % list_files)
    data = {}

    if list_files:
        var2load = [
            "lwp",
            "Z",
            "beta",
            "mean_zbeta",
            "temperature",
            "v",
            "pressure",
            "specific_humidity",
            "rainrate",
        ]

        # open all files
        nc_ids = [nc.Dataset(file_) for file_ in list_files]

        # localization of instrument
        data["lat"] = nc_ids[0].variables["latitude"][:]
        data["lon"] = nc_ids[0].variables["longitude"][:]
        data["alt"] = nc_ids[0].variables["altitude"][:]
        data["location"] = nc_ids[0].location

        # read alt (no need to concatenate)
        data["height"] = nc_ids[0].variables["height"][:]
        data["model_height"] = nc_ids[0].variables["model_height"][:]

        # wavelength (no need to concatenate)
        tmp = nc_ids[0].variables["lidar_wavelength"][:]
        data["wavelength"] = tmp
        data["wavelength_units"] = nc_ids[0].variables["lidar_wavelength"].units

        # read time
        units = nc_ids[0].variables["time"].units  # HOURS SINCE %Y-%m-%d 00:00:00 +0:00
        tmp = [nc.num2date(nc_id.variables["time"][:], units) for nc_id in nc_ids]
        data["raw_time"] = np.concatenate(tmp)

        # check if any data available
        # print(date)
        # time_filter = (data['raw_time'] >= date) & (data['raw_time'] < date + dt.timedelta(days=1))
        # if not np.any(time_filter):
        #     return None

        for var in var2load:
            tmp = [nc_id.variables[var][:] for nc_id in nc_ids]
            data[var] = np.ma.filled(np.concatenate(tmp, axis=0), fill_value=np.nan)
            # It is assumed that properties do not change along netcdf files for the same day
            data[var][data[var] == nc_ids[0][var].missing_value] = np.nan

        # Change name to fit real array content
        data["dBZe"] = data["Z"].copy()

        # close all files
        [nc_id.close() for nc_id in nc_ids]

    return data


def classification_reader(list_files: list[str]) -> dict:
    """read data from netCDF files

    Designed to work with netcdf list_files associated to one single day
    """
    """
    0: Clear sky
    1: Cloud liquid droplets only
    2: Drizzle or rain
    3: Drizzle or rain coexisting with cloud liquid droplets
    4: Ice particles
    5: Ice coexisting with supercooled liquid droplets
    6: Melting ice particles
    7: Melting ice particles coexisting with cloud liquid droplets
    8: Aerosol particles, no cloud or precipitation
    9: Insects, no cloud or precipitation
    10: Aerosol coexisting with insects, no cloud or precipitation
    """
    logger.warning("to be deprecated")
    print("Reading netCDF files: %s" % list_files)

    data = {}

    if list_files:
        var2load = ["cloud_base_height", "cloud_top_height", "target_classification"]
        # open all files
        nc_ids = [nc.Dataset(file_) for file_ in list_files]

        # localization of instrument
        data["lat"] = nc_ids[0].variables["latitude"][:]
        data["lon"] = nc_ids[0].variables["longitude"][:]
        data["alt"] = nc_ids[0].variables["altitude"][:]
        data["location"] = nc_ids[0].location

        # read alt (no need to concantenate)
        data["height"] = nc_ids[0].variables["height"][:]

        # read time
        units = (
            nc_ids[0].variables["time"].units
        )  #'days since %s' % dt.datetime.strftime(date, '%Y-%m-%d %H:%M:%S')
        tmp = [nc.num2date(nc_id.variables["time"][:], units) for nc_id in nc_ids]
        data["raw_time"] = np.concatenate(tmp)

        # check if any data available
        # print(date)
        # time_filter = (data['raw_time'] >= date) & (data['raw_time'] < date + dt.timedelta(days=1))
        # if not np.any(time_filter):
        #     return None

        for var in var2load:
            tmp = [nc_id.variables[var][:] for nc_id in nc_ids]
            data[var] = np.ma.filled(np.concatenate(tmp, axis=0), fill_value=np.nan)
            # It is assumed that properties do not change along netcdf files for the same day
            if "missing_value" in nc_ids[0][var].ncattrs():
                data[var][data[var] == nc_ids[0][var].missing_value] = np.nan

        # close all files
        [nc_id.close() for nc_id in nc_ids]

    return data



""" Retrievals of backscatter and extinction based on Raman measurements

.. warning::
   These functions have not been tested!
"""
from pdb import set_trace
from typing import Tuple
import numpy as np
from scipy.integrate import cumtrapz, cumulative_trapezoid
from scipy.signal import savgol_filter

from gfatpy.atmo.atmo import molecular_properties, number_density_at_pt
from gfatpy.lidar.utils import signal_to_rcs


def retrieve_extinction(
    signal: np.ndarray,
    ranges: np.ndarray,
    wavelengths: Tuple[float, float],
    pressure: np.ndarray,
    temperature: np.ndarray,
    window_size_m: float = 100.0,
    savgol_order: int = 1,
    particle_angstrom_exponent: float = 1,
) -> tuple[np.ndarray, None]:
   """Retrieve particle extinction from lidar signal.

   This function calculates the particle extinction from the lidar signal using
   range corrected signal (rcs) and molecular properties.

   Args:
      signal (np.ndarray): Lidar signal.
      ranges (np.ndarray): Array of range values corresponding to the signal.
      wavelengths (Tuple[float, float]): Tuple of two wavelengths (in nm).
      pressure (np.ndarray): Atmospheric pressure profile.
      temperature (np.ndarray): Atmospheric temperature profile.
      window_size_m (float, optional): Window size for derivative calculation in meters.
      savgol_order (int, optional): Order of Savitzky-Golay filter.
      particle_angstrom_exponent (float, optional): Particle Ångström exponent.

   Returns:
      Union[np.ndarray, None]: Array containing the particle extinction values. 
      If the number density or signal is invalid, the corresponding element in 
      the result will be set to np.nan. If the input signal and number density 
      are both invalid, the function returns None.
   """

   #Retrieve range corrected signal
   rcs = signal_to_rcs(signal, ranges)

   #Retrieve molecular extinction
   atmo_data = molecular_properties(wavelengths[0], pressure, temperature, heights=ranges)
   elastic_alpha_molecular = atmo_data["molecular_alpha"].values
   atmo_data_raman  = molecular_properties(wavelengths[1], pressure, temperature, heights=ranges)   
   raman_alpha_molecular = atmo_data_raman ["molecular_alpha"].values
   
   #Retrieve number density
   number_density = number_density_at_pt(pressure, temperature)


   #Calculate ratio
   ratio = np.nan * np.ones(signal.shape)
   valid_idx = np.logical_and(number_density > 0, signal > 0)
   ratio[valid_idx] = np.ma.log( number_density[valid_idx] / rcs[valid_idx]  )

   # Apply derivative
   dz = np.median(np.diff(ranges))  # type: ignore 
   window_size_bin = np.floor(window_size_m / dz).astype(int)
   derivative = savgol_filter(ratio, window_size_bin, savgol_order, deriv=1, delta=dz, mode="nearest", cval=np.nan)  # Calculate 1st derivative

   # Calculate particle extinction
   cte = 1 + (wavelengths[1]/wavelengths[0])**(-particle_angstrom_exponent)
   particle_alpha = (derivative - raman_alpha_molecular - elastic_alpha_molecular)/cte
   
   #Fill like overlap
   idx_overlap = np.ceil(window_size_bin/2).astype(int)
   particle_alpha[:idx_overlap] = particle_alpha[idx_overlap]

   return particle_alpha


def retrieve_backscatter(
    signal_raman: np.ndarray,
    signal_emission: np.ndarray,
    extinction_profile: np.ndarray,
    range_profile: np.ndarray,
    wavelengths: Tuple[float, float],
    pressure: np.ndarray,
    temperature: np.ndarray,
    particle_angstrom_exponent: float = 1,
    reference: Tuple[float, float] = (8000, 8500),
    beta_aer_ref: float = 0,
) -> np.ndarray:
   """Calculates the aerosol backscatter coefficient.

   This function calculates the aerosol backscatter coefficient based on preprocessed 
   elastic and Raman signals and the retrieved aerosol extinction coefficient.

   Parameters:
      signal_raman (np.ndarray): The range-corrected Raman signal (1D array of size M).
      signal_emission (np.ndarray): The range-corrected elastic signal (at the emission wavelength, 1D array of size M).
      extinction_profile (np.ndarray): The aerosol extinction coefficient at each point of the signal profile (1D array of size M).
      range_profile (np.ndarray): Array of range values corresponding to the signal (1D array of size M).
      wavelengths (Tuple[float, float]): Tuple of two wavelengths (in nm) for the emission and Raman signals.
      pressure (np.ndarray): Atmospheric pressure profile (1D array of size M, [Pa]).
      temperature (np.ndarray): Atmospheric temperature profile (1D array of size M, [K]).
      particle_angstrom_exponent (float, optional): Particle Ångström exponent (default: 1).
      reference (Tuple[float, float], optional): Reference altitude range (default: (8000, 8500)).
      beta_aer_ref (float, optional): The molecular backscatter coefficient at the reference altitude (default: 0).

   Returns:
      np.ndarray: The aerosol backscatter coefficient [m^-1].

   Notes:
      The aerosol backscatter coefficient is given by the formula:
      β_aer(R, λ0) = [β_aer(R0, λ0) + β_mol(R0, λ0)] * (P(R0, λ_ra) * P(R, λ0)) / (P(R0, λ0) * P(R, λ_ra)) *
                     exp(-∫(R0, R) [α_aer(r, λ_ra) + α_mol(r, λ_ra)] dr) / exp(-∫(R0, R) [α_aer(r, λ0) + α_mol(r, λ0)] dr) -
                     β_mol(R, λ0)

   References:
      Ansmann, A. et al. Independent measurement of extinction and backscatter profiles in cirrus clouds by using a combined Raman elastic-backscatter lidar.
      Applied Optics Vol. 31, Issue 33, pp. 7113-7131 (1992)
   """
   idx_ref = np.logical_and(range_profile >= reference[0], range_profile <= reference[1])
   if not idx_ref.any():
      raise ValueError("Range `reference` out of rcs size.")

   # Calculate profiles of molecular extinction
   atmo_data = molecular_properties(wavelengths[0], pressure, temperature, heights=range_profile)
   beta_mol = atmo_data["molecular_beta"].values
   alpha_mol_emmision = atmo_data["molecular_alpha"].values
   atmo_data_raman = molecular_properties(wavelengths[1], pressure, temperature, heights=range_profile)
   beta_mol_raman = atmo_data_raman["molecular_beta"].values
   alpha_mol_raman = atmo_data_raman["molecular_alpha"].values

   
   alpha_part_raman = extinction_profile * (wavelengths[1] / wavelengths[0]) ** (-particle_angstrom_exponent ) 

   T_raman = np.exp(-cumulative_trapezoid(alpha_mol_raman + alpha_part_raman, range_profile, initial=0))  # type: ignore
   T_elastic = np.exp(-cumulative_trapezoid(alpha_mol_emmision + extinction_profile, range_profile, initial=0))  # type: ignore

   #Calibration at reference altitude
   beta_part_ref = 0
   beta_at_reference = beta_part_ref + np.nanmean(beta_mol[idx_ref])
   gain_ratio = (signal_emission / signal_raman) * (T_raman/T_elastic) * (beta_mol_raman/beta_at_reference)
   gain_ratio_at_reference = np.nanmean(gain_ratio[idx_ref])
   beta_part = beta_mol_raman*(1/gain_ratio_at_reference)*((signal_emission * T_raman)/(signal_raman*T_elastic)) - beta_mol
   return beta_part



import numpy as np
from numpy.polynomial.polynomial import polyfit

import numba


def best_slope_fit(mat1: np.ndarray, mat2: np.ndarray, window: int) -> np.ndarray:

    assert mat1.shape == mat2.shape, "Matrices shape must match"
    assert isinstance(window, int), "Window argument must be an integer"

    x = np.arange(window)
    res = np.array([])

    # Iter in first dimension, equivalent to time in lidar case
    for idx in range(mat1.shape[0]):
        windowed1 = rolling(mat1[idx], window)
        windowed2 = rolling(mat2[idx], window)

        slopes1 = polyfit(x, windowed1.T, 1)[0]
        slopes2 = polyfit(x, windowed2.T, 1)[0]
        chosen_group = np.argmin(np.abs(slopes1 - slopes2) / slopes1)
        res = np.hstack([res, chosen_group + np.floor(window / 2)])  # Append chosen bin

    return res.astype(int)


@numba.njit(parallel=True)
def windowed_corrcoefs(arr1: np.ndarray, arr2: np.ndarray, w_size: int):
    range_shape = arr1.shape[1] - (w_size - 1)
    _corrcoefs = np.empty((arr1.shape[0], range_shape))
    for t_idx in numba.prange(arr1.shape[0]):
        w1 = rolling(arr1[t_idx], w_size)
        w2 = rolling(arr2[t_idx], w_size)
        for idx in numba.prange(range_shape):
            _w1 = w1[idx]
            _w2 = w2[idx]
            coeff = np.corrcoef(_w1, _w2)[1, 0]
            _corrcoefs[t_idx][idx] = coeff
    return _corrcoefs


@numba.njit(parallel=True)
def windowed_proportional(arr1: np.ndarray, arr2: np.ndarray, /, *, w_size: int):

    assert arr1.shape == arr2.shape, "Matrices shape must match"
    # assert isinstance(w_size, int), "Window argument must be an integer"

    range_shape = arr1.shape[1] - (w_size - 1)
    _proportional = np.full((arr1.shape[0], range_shape), np.nan)
    _factor = np.full((arr1.shape[0], range_shape), np.nan)

    for t_idx in numba.prange(arr1.shape[0]):
        w1 = rolling(arr1[t_idx], w_size)
        w2 = rolling(arr2[t_idx], w_size)

        for idx in numba.prange(range_shape):
            _w1 = w1[idx]
            _w2 = w2[idx]

            ratio = np.mean(_w2 / _w1)
            adj = _w1 * ratio
            _factor[t_idx][idx] = ratio
            _proportional[t_idx][idx] = (np.abs(adj - _w2) / _w2).mean()

    return _factor, _proportional


@numba.njit()
def rolling(a, window):
    shape = (a.size - window + 1, window)
    strides = (a.itemsize, a.itemsize)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


@numba.njit
def correlate_vector_to_matrix(vector, matrix):
    """Correlate a vector to a matrix

    Args:
        vector (np.ndarray): Vector to correlate
        matrix (np.ndarray): Matrix to correlate

    Returns:
        np.ndarray: Correlation result
    """
    corr = np.zeros(matrix.shape[1])
    for i in range(matrix.shape[1]):
        corr[i] = np.corrcoef(vector, matrix[:, i])[0, 1]
    return corr



import sys
from typing import Literal

from loguru import logger


def set_logger_level(
    level: Literal["TRACE", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
) -> None:
    logger.remove()
    logger.add(sys.stderr, level=level)



from typing import Callable
from warnings import warn


import numpy as np
import xarray as xr
from scipy.signal import savgol_filter

from gfatpy.utils.optimized import best_slope_fit


def _create_smoother_normalizer(window_size: int) -> Callable[[np.ndarray], np.ndarray]:
    def smooth_and_normalize(row: np.ndarray) -> np.ndarray:
        return savgol_filter(row / np.nanmean(row), window_size, 4)

    return smooth_and_normalize


# # TODO: This part can be paralellized with Numba
# def get_gluing_indexes(
#     signal_an: np.ndarray, rcs_pc: np.ndarray, half_window_bins: int
# ) -> None:

#     # To paralelize with Numba
#     for time_idx in np.arange(signal_an.shape[0]):
#         for h_idx in np.arange(rcs_an.shape[1]):
#             ...


def gluing(
    signal_an: xr.DataArray,
    signal_pc: xr.DataArray,
    range_min: float = 1500,
    range_max: float = 5000,
    criteria_half_window: float = 300,
    adjustment_half_window: float = 2200,
):

    sel_signal_an = signal_an.sel(
        range=slice(range_min - criteria_half_window, range_max + criteria_half_window)
    )
    sel_signal_pc = signal_pc.sel(
        range=slice(range_min - criteria_half_window, range_max + criteria_half_window)
    )

    sel_range = sel_signal_pc.range.values
    window_size_bins = int((criteria_half_window * 2) / (sel_range[1] - sel_range[0]))

    smooth_window = 11  # window_size_bins // 2

    sm_sel_signal_an = np.apply_along_axis(
        _create_smoother_normalizer(smooth_window), 0, sel_signal_an.values
    )
    sm_sel_signal_pc = np.apply_along_axis(
        _create_smoother_normalizer(smooth_window), 0, sel_signal_pc.values
    )

    signal_glued = signal_pc.values.copy()

    gluing_bins = best_slope_fit(
        sm_sel_signal_an, sm_sel_signal_pc, window=window_size_bins
    )
    gluing_ranges = sel_range[gluing_bins]

    for idx, gluing_range in enumerate(gluing_ranges[:]):
        near_idx = (sel_range > (gluing_range - adjustment_half_window)) & (
            sel_range < (gluing_range + adjustment_half_window)
        )

        sel_an = sel_signal_an[dict(time=idx)][near_idx]
        sel_pc = sel_signal_pc[dict(time=idx)][near_idx]

        r_gluing = np.corrcoef(sm_sel_signal_an[idx], sm_sel_signal_pc[idx])[0, 1]

        fit_values = np.polyfit(
            sel_an,
            sel_pc,
            1,
        )

        if r_gluing < 1:  # TODO: Adjust threshold
            signal_an2pc = np.polyval(fit_values, signal_an.values[idx, :])
            signal_glued[idx, :] = np.concatenate(
                [
                    signal_an2pc[signal_pc.range.values < gluing_range],
                    signal_pc[idx, signal_pc.range.values >= gluing_range],
                ]
            )
        else:
            warn(
                f"r2 thereshold condition not satisfied for profile in time index {idx}"
            )
            # TODO: warn or implement filling method

    return signal_glued



#!/usr/bin/env python
import os
import glob
import pathlib
import warnings

import datetime as dt
import numpy as np
import xarray as xr
from loguru import logger

from gfatpy import utils
from gfatpy.radar.utils import RADAR_INFO


warnings.filterwarnings("ignore")


__author__ = "Bravo-Aranda, Juan Antonio"
__license__ = "GPL"
__version__ = "1.0"
__maintainer__ = "Bravo-Aranda, Juan Antonio"
__email__ = "jabravo@ugr.es"
__status__ = "Production"

""" DEFAULT AUXILIAR INFO
"""

def reader_xarray(
    filelist: list[str] | str,
    properties: list[str] = [],
) -> xr.Dataset:

    def select_properties(dataset: xr.Dataset, properties: list | str) -> xr.Dataset:
        """select_properties function creates a new dataset with 'properties' (list).

        Args:
            dataset (xr.Dataset): radar dataset
            properties (list | str): list of radar property names

        Returns:
            xr.Dataset: radar dataset
        """

        if len(properties) > 0:
            if isinstance(properties, str):
                properties = [properties]
            if isinstance(properties, np.ndarray):
                properties = properties.tolist()

            # find properties
            for _property in [*dataset.variables.keys()]:
                if _property not in properties:                                            
                    dataset = dataset.drop_vars(_property)                
        return dataset

    #load the files
    radar = None
    for fn in filelist:
        with xr.open_dataset(
            fn, chunks={}
        ) as _dx:
            _dx = select_properties(_dx, properties)
        if not radar:
            radar = _dx
        else:
            # concat only variables that have "time" dimension.
            # rest of variables keep values from first dataset
            try:
                radar = xr.concat(
                    [radar, _dx],
                    dim="time",
                    data_vars="minimal",
                    coords="minimal",
                    compat="override",
                )
            except Exception as e:
                logger.critical(f"Dataset in {fn} not concatenated")
                raise e
    return radar


from enum import Enum
from typing import TypedDict

from typing_extensions import NotRequired

import numpy as np


class DataType(str, Enum):
    raw = "RAW"
    produt = "PRODUCTS"

    def __str__(self):
        return self.value

    def to_level(self):
        if self == DataType.raw:
            return "0a"
        elif self == DataType.produt:
            return "1a"
        else:
            raise ValueError("Invalid DataType")


class AgoraStation(str, Enum):
    ugr = "UGR"
    sns = "SNS"

    def __str__(self):
        return self.value


class LidarName(str, Enum):
    mhc = "mulhacen"
    alh = "alhambra"
    vlt = "veleta"

    def __str__(self):
        return self.value


class MeasurementType(str, Enum):
    RS = "RS"
    HF = "HF"
    DC = "DC"
    TC = "TC"
    DP = "DP"
    OT = "OT"

    def __str__(self):
        return self.value


class Telescope(str, Enum):
    xf = "xf"
    ff = "ff"
    nf = "nf"

    def __str__(self):
        return self.value


class _LidarInfo(TypedDict):
    mulhacen: dict
    alhambra: dict
    veleta: dict


class _LidarMetadata(TypedDict):
    nick2name: dict
    name2nick: dict
    measurement_type: list
    code_telescope_str2number: dict
    code_mode_str2number: dict
    code_polarization_str2number: dict
    code_mode_number2str: dict
    code_polarization_number2str: dict


class LidarInfoType(TypedDict):
    lidars: _LidarInfo
    metadata: _LidarMetadata


class ParamsDict(TypedDict):
    k_lidar: float
    particle_alpha: np.ndarray
    particle_alpha_raman: NotRequired[np.ndarray]
    particle_beta: np.ndarray
    particle_beta_raman: NotRequired[np.ndarray]
    molecular_alpha: np.ndarray
    molecular_alpha_raman: NotRequired[np.ndarray]
    molecular_beta: np.ndarray
    molecular_beta_raman: NotRequired[np.ndarray]
    particle_accum_ext: np.ndarray
    particle_accum_ext_raman: NotRequired[np.ndarray]
    molecular_accum_ext: np.ndarray
    molecular_accum_ext_raman: NotRequired[np.ndarray]
    molecular_beta_att: np.ndarray
    transmittance: np.ndarray
    overlap: np.ndarray
    angstrom_exponent_fine: float
    angstrom_exponent_coarse: float



from pdb import set_trace
import numpy as np
import xarray as xr

from gfatpy.utils.optimized import windowed_proportional


def gluing(
    signal_an: xr.DataArray,
    signal_pc: xr.DataArray,
    range_min: float = 1100,
    range_max: float = 5000,
    window_meters: float = 500,
    # max_rel_error: float = 0.10,
):

    t_slice = slice(range_min - window_meters / 2, range_max + window_meters / 2)

    rcs_an = signal_an.sel(range=t_slice)
    rcs_an = rcs_an * signal_an.range**2

    rcs_pc = signal_pc.sel(range=t_slice)
    rcs_pc = rcs_pc * signal_pc.range**2

    ranges = rcs_an.range.values
    ranges_complete = signal_an.range.values
    window_size = int(window_meters // (ranges[1] - ranges[0]))

    coeff, error = windowed_proportional(
        rcs_an.values, rcs_pc.values, w_size=window_size
    )

    min_error = np.argmin(error, axis=1)
    gluing_height = signal_an.sel(range=slice(range_min, range_max)).range.values[
        min_error
    ]
    # if len(gluing_height.shape) > 0:
    #     gluing_height = gluing_height[0]

    # TODO: Coefs an be overwritten and smoothen along time axis with savgol filter
    # coeff = coeff[:][min_error]

    rcs_gl = (signal_pc * signal_pc.range**2).values.copy()
    an_values = (signal_an * signal_an.range**2).values

    for time_idx in np.arange(signal_an.time.shape[0]):
        mask = ranges_complete < gluing_height[time_idx]
        # set_trace()
        rcs_gl[time_idx][mask] = (
            an_values[time_idx][mask] * coeff[time_idx][min_error[time_idx]]
        )

        # if error[time_idx][min_error[time_idx]] > max_rel_error:
        #     rcs_gl[time_idx] = np.nan

    return signal_an.copy(data=rcs_gl).rename(signal_an.name[:-1] + "g"), error, gluing_height  # type: ignore

    # rolling = sliding_window_view(np.arange(ranges.shape[0]), int(window_meters // window_size))

    # # xr.full_like(signal_an.sel(range=slice(range_min, range_max)), np.nan)

    # # np.empty((signal_an.shape[0], ranges.shape[0]))

    # for idx, window in enumerate(rolling):
    #     an_window = rcs_an.values.T[window].T
    #     pc_window = rcs_pc.values.T[window].T

    #     ratio = (pc_window / an_window).mean(axis=0)
    #     an_adj = an_window * ratio

    #     error: np.ndarray = np.abs(an_adj - pc_window) / pc_window  # type: ignore
    #     error.mean(axis=1)

    #     # set_trace()

    #     errors[idx] = error.mean(axis=1)
    #     set_trace()
    #     # print(f"Errors calculated for {_range}")

    # return errors


# @nb.njit(parallel = True)
# def evaluate_errors(an: np.ndarray, pc: np.ndarray) -> np.ndarray:


#     set_trace()
